from keras.src.backend.config import backend
from keras.src.backend import any_symbolic_tensors
from keras.src.ops.operation import Operation
from keras import ops
from keras import KerasTensor
from typing import Union, Tuple

if backend() == 'jax':
    from .jax import fft_fn, ifft_fn, fft2_fn, ifft2_fn, fft3_fn, ifft3_fn, rfft_fn, irfft_fn, rfft2_fn, irfft2_fn, rfft3_fn, irfft3_fn
    from .jax import fft_derivative as _fft_derivative

if backend() == 'tensorflow':
    from .tensorflow import fft_fn, ifft_fn, fft2_fn, fft3_fn, ifft3_fn, ifft2_fn, rfft_fn, irfft_fn, rfft2_fn, irfft2_fn, rfft3_fn, irfft3_fn
    from .tensorflow import fft_derivative as _fft_derivative

def cast_to_complex(x: Union[Tuple[KerasTensor, KerasTensor], KerasTensor]) -> Tuple[KerasTensor, KerasTensor]:
    if isinstance(x, tuple):
        return x
    else:
        return x, ops.zeros_like(x, dtype=x.dtype)


# === FFT ===
class FFT(Operation):
    """
    Base 1-D fast Fourier transform (FFT) operation.
    Keras-backend-agnostic version of the FFT.

    Notes
    -----
    Keras3 does not support complex dtypes.
    Therefore, the `keras.ops.fft` is quite cumbersume to use, since it explicitly requires the definition of a real and a complex part when calling `fft` ::

    >>> from keras import ops
    >>> x = ops.ones((4,))
    >>> ops.fft(x)
    ValueError: The truth value of an array with more than one element is ambiguous. Use a.any() or a.all()
    >>> ops.fft((x, ops.zeros_like(x)))
    array([4.,  0.,  0.,  0.], dtype=float32), array([0., 0., 0., 0.], dtype=float32)
    
    This implementation attempts to increase the user-friendliness of the FFT call in Keras by handling the input automatically.

    """

    def __init__(self):
        super().__init__()
        self.fft_fn = fft_fn

    def compute_output_spec(self, x):
        """
        Compute output spec of Fourier transform

        Parameters
        ----------
        x : KerasTensor | tuple | list
            Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

        Returns
        -------
        y_real_spec, y_imag_spec : (KerasTensor, KerasTensor)
            spec of real- and imaginary part of `FFT(x)`

        """

        if not isinstance(x, (tuple, list)) or len(x) != 2:
            real = x
            imag = ops.zeros_like(x)
        else:
            real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )

        return (
            KerasTensor(shape=real.shape, dtype=real.dtype),
            KerasTensor(shape=imag.shape, dtype=imag.dtype),
        )

    def call(self, x):
        """
        Call method of FFT

        Parameters
        ----------
        x : KerasTensor | tuple | list
            Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

        Returns
        -------
        y_real, y_imag : (KerasTensor, KerasTensor)
            real- and imaginary part of `FFT(x)`

        Notes
        -----
        The input can be either a KerasTensor or a tuple or list of length 2, containing the real- and imaginary part of `x`.

        """

        return self.fft_fn(x)
    

class FFT2(FFT):
    """
    2-D fast Fourier transform

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = fft2_fn


class FFT3(FFT):
    """
    3-D fast Fourier transform

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = fft3_fn


class RFFT(FFT):
    """
    1-D fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = rfft_fn


class RFFT2(FFT):
    """
    2-D fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = rfft2_fn


class RFFT3(FFT):
    """
    3-D fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = rfft3_fn


# === IFFT ===
class IFFT(FFT):
    """
    1-D inverse fast Fourier transform

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = ifft_fn


class IFFT2(FFT):
    """
    2-D inverse fast Fourier transform

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = ifft2_fn


class IFFT3(FFT):
    """
    3-D inverse fast Fourier transform

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = ifft3_fn


class IRFFT(FFT):
    """
    1-D inverse fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = irfft_fn

    def compute_output_spec(self, x):
        """
        Compute output spec of Fourier transform

        Parameters
        ----------
        x : KerasTensor | tuple | list
            Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

        Returns
        -------
        y_real_spec : KerasTensor
            spec of real part of `IRFFT(x)`

        """

        if not isinstance(x, (tuple, list)) or len(x) != 2:
            real = x
            imag = ops.zeros_like(x)
        else:
            real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )

        return KerasTensor(shape=real.shape, dtype=real.dtype)


class IRFFT2(IRFFT):
    """
    2-D inverse fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = irfft2_fn


class IRFFT3(IRFFT):
    """
    3-D inverse fast Fourier transform for real-valued inputs

    """

    def __init__(self):
        super.__init__()
        self.fft_fn = irfft3_fn


# === now make function wrapper ===
# === forward FFT ===
def fft(x):
    """
    1-D fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of FFT(x).

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.array([-2, 8, 6, 4, 1, 0, 3, 5], dtype=float)
    >>> y_real, y_imag = fft(signal)
    >>> ops.convert_to_numpy(y_real)
    array([ 25.       ,   3.3639607, -10.       ,  -9.36396  ,  -9.       ,
        -9.36396  , -10.       ,   3.3639607], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([ 0.       , -7.9497476,  1.       , -1.9497476,  0.       ,
        1.9497476, -1.       ,  7.9497476], dtype=float32)

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return FFT().symbolic_call(x)
    return fft_fn(x)


def fft2(x):
    """
    2-D fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of FFT2(x).

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.reshape(ops.array([-2, 8, 6, 12], dtype=float), (2, 2))
    >>> y_real, y_imag = fft2(signal)
    >>> ops.convert_to_numpy(y_real)
    array([
        [ 16.,  -8.],
        [ -4., -12.]
    ], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([
        [0., 0.],
        [0., 0.]
    ], dtype=float32)

    """
    
    if any_symbolic_tensors(cast_to_complex(x)):
        return FFT2().symbolic_call(x)
    return fft2_fn(x)


def fft3(x):
    """
    3-D fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of FFT3(x).

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.reshape(ops.array([-2, 8, 6, 12], dtype=float), (2, 2))
    >>> y_real, y_imag = fft3(signal)
    >>> ops.convert_to_numpy(y_real)
    array([
        [ 16.,  -8.],
        [ -4., -12.]
    ], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([
        [0., 0.],
        [0., 0.]
    ], dtype=float32)

    """
    
    if any_symbolic_tensors(cast_to_complex(x)):
        return FFT3().symbolic_call(x)
    return fft3_fn(x)


def rfft(x):
    """
    1-D fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor
        Real-valued input to FFT.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of RFFT(x).
        Both have size (n // 1 + 1) given ops.shape(x) == n

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.array([-2, 8, 6, 4, 1, 0, 3, 5], dtype=float)
    >>> y_real, y_imag = rfft(signal)
    >>> ops.convert_to_numpy(y_real)
    array([ 25.       ,   3.3639607, -10.       ,  -9.36396  ,  -9.       ], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([ 0.       , -7.9497476,  1.       , -1.9497476,  0.       ], dtype=float32)
       
    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return RFFT().symbolic_call(x)
    return rfft_fn(x)


def rfft2(x):
    """
    2-D fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor
        Real-valued input to FFT.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of RFFT2(x).
        Both have size (n, n // 1 + 1) given ops.shape(x) == (n, n)

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.reshape(ops.array([-2, 8, 6, 12, 4, 0, 2, 2, -2, 8, 6, 12, 4, 0, 2, 2], dtype=float), (4, 4))
    >>> y_real, y_imag = rfft2(signal)
    >>> ops.convert_to_numpy(y_real)
    array([
        [ 64., -12., -24.],
        [  0.,   0.,   0.],
        [ 32., -20., -40.],
        [  0.,   0.,   0.]
    ], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([
        [ 0., 12.,  0.],
        [ 0.,  0.,  0.],
        [ 0.,  4.,  0.],
        [ 0.,  0.,  0.]
    ], dtype=float32)

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return RFFT2().symbolic_call(x)
    return rfft2_fn(x)


def rfft3(x):
    """
    3-D fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor
        Real-valued input to FFT.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of RFFT3(x).
        Both have size (n, n // 1 + 1) given ops.shape(x) == (n, n)

    Examples
    --------
    >>> from keras import ops
    >>> signal = ops.reshape(ops.array([-2, 8, 6, 12, 4, 0, 2, 2, -2, 8, 6, 12, 4, 0, 2, 2], dtype=float), (4, 4))
    >>> y_real, y_imag = rfft3(signal)
    >>> ops.convert_to_numpy(y_real)
    array([
        [ 64., -12., -24.],
        [  0.,   0.,   0.],
        [ 32., -20., -40.],
        [  0.,   0.,   0.]
    ], dtype=float32)
    >>> ops.convert_to_numpy(y_imag)
    array([
        [ 0., 12.,  0.],
        [ 0.,  0.,  0.],
        [ 0.,  4.,  0.],
        [ 0.,  0.,  0.]
    ], dtype=float32)

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return RFFT3().symbolic_call(x)
    return rfft3_fn(x)


# === inverse FFT ===
def ifft(x):
    """
    1-D inverse fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IFFT(x).

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return IFFT().symbolic_call(x)
    return ifft_fn(x)


def ifft2(x):
    """
    2-D inverse fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IFFT2(x).

    """
    
    if any_symbolic_tensors(cast_to_complex(x)):
        return IFFT2().symbolic_call(x)
    return ifft2_fn(x)


def ifft3(x):
    """
    3-D inverse fast Fourier transform

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IFFT3(x).

    """
    
    if any_symbolic_tensors(cast_to_complex(x)):
        return IFFT3().symbolic_call(x)
    return ifft3_fn(x)


def irfft(x, n=None):
    """
    1-D inverse fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.
    n : int, optional
        Length of the transformed axis of the output. For `n` output points, `n//2+1` input points are necessary.
        If the input is longer than tis, it is cropped.
        If it is shorter than this, it is padded with zeros.
        If `n` is not given, it is taken to be `2`(m-1)` where `m` is the length of the input along the axis specified by axis,
        cf. https://numpy.org/doc/stable/reference/generated/numpy.fft.irfft.html

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IRFFT(x).
        Both have size (n // 1 + 1) given ops.shape(x) == n

    """
    
    if any_symbolic_tensors(cast_to_complex(x)):
        return IRFFT().symbolic_call(x, n=n)
    return irfft_fn(x, n=n)


def irfft2(x, n=None):
    """
    2-D inverse fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.
    n : tuple, optional
        Shape (length of each transformed axis) of the output.
        `s` is also the number of input points used along this axis, except for the last axis,
        where `s[-1]//2+1` points of the input are used.
        Along any axis, if the shape indicated by `s` is smaller than that of the input, the input is cropped.
        If it is larger, the input is padded with zeros,
        cf. https://numpy.org/doc/stable/reference/generated/numpy.fft.irfftn.html
        
    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IRFFT2(x).
        Both have size (n // 1 + 1, n) given ops.shape(x) == (n, n)

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return IRFFT2().symbolic_call(x, n=n)
    return irfft2_fn(x, n=n)


def irfft3(x, n=None):
    """
    3-D inverse fast Fourier transform for real-valued inputs

    Parameters
    ----------
    x : KerasTensor | tuple | list
        Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.
    n : tuple, optional
        Shape (length of each transformed axis) of the output.
        `s` is also the number of input points used along this axis, except for the last axis,
        where `s[-1]//2+1` points of the input are used.
        Along any axis, if the shape indicated by `s` is smaller than that of the input, the input is cropped.
        If it is larger, the input is padded with zeros,
        cf. https://numpy.org/doc/stable/reference/generated/numpy.fft.irfftn.html

    Returns
    -------
    y_real, y_imag : (KerasTensor, KerasTensor)
        Tuple of real- and imaginary part of IRFFT3(x).
        Both have size (n // 1 + 1, n) given ops.shape(x) == (n, n)

    """

    if any_symbolic_tensors(cast_to_complex(x)):
        return IRFFT3().symbolic_call(x, n=n)
    return irfft3_fn(x, n=n)


# === FFT Derivative ===
class FFTDerivative(Operation):
    def __init__(self):
        super().__init__()

    def compute_output_spec(self, x):
        """
        Compute output spec of Fourier transform

        Parameters
        ----------
        x : KerasTensor | tuple | list
            Real- or complex input to FFT. A complex input must be composed of a tuple or list of the real- and imaginary part `(x_real, x_imag)`.

        Returns
        -------
        diff_spec : KerasTensor
            spec of derivative of x

        """
        if not isinstance(x, (tuple, list)) or len(x) != 2:
            real = x
            imag = ops.zeros_like(x)
        else:
            real, imag = x
        # Both real and imaginary parts should have the same shape.
        if real.shape != imag.shape:
            raise ValueError(
                "Input `x` should be a tuple of two tensors - real and "
                "imaginary. Both the real and imaginary parts should have the "
                f"same shape. Received: x[0].shape = {real.shape}, "
                f"x[1].shape = {imag.shape}"
            )
        return KerasTensor(shape=real.shape, dtype=real.dtype)
    
    def call(self, x, axis=-1, d=1.0, n=1):
        """
        Derivative in Fourier space

        Parameters
        ----------
        x : tf.Tensor
            Input array, can be either real or complex valued.
        axis : int (optional)
            Axis along which `x` is derived.
            Defaults to `-1`.
        d : float (optional)
            Discretization of `x` along the dimension of `axis`.
            Defaults to `1.0`.
        n : int (optional)
            Order of differentiation.
            Defaults to `1`.

        """

        return _fft_derivative(x=x, axis=axis, d=d, n=n)
    
def fft_derivative(x, axis=-1, d=1.0, n=1):
    """
    Derivative in Fourier space

    Parameters
    ----------
    x : tf.Tensor
        Input array, can be either real or complex valued.
    axis : int (optional)
        Axis along which `x` is derived.
        Defaults to `-1`.
    d : float (optional)
        Discretization of `x` along the dimension of `axis`.
        Defaults to `1.0`.
    n : int (optional)
        Order of differentiation.
        Defaults to `1`.

    """
    if any_symbolic_tensors(cast_to_complex(x)):
        return FFTDerivative.symbolic_call(x)
    return fft_derivative(x)



from keras import models
from keras import regularizers, initializers, constraints
from keras import saving
from ...ops.helper import _IterableVars
from importlib import import_module


class BaseFCN(models.Model, _IterableVars):
    """
    Base class of FullyConvolutionalNetwork (FCN)

    Convolutional autoencoder *without* skip-connections

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseFCN`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional
        Padding for encoder blocks. 
        Decoder blocks will mirror the `padding` of the encoder blocks.
        If `rank=1`, `padding` may be either `"same"` or `"causal"`, `rank>1` enforces `padding="same"`
        to maintain deterministic shapes throughout the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    use_skip_connection : bool | list, optional
        Whether to use skip connections (basically transform FCN to a Unet).
        Can be defined per layer, i.e., 
        `use_skip_connection=[True, False, False]` results in a skip connection on the highest level only.
        Defaults to `False`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from skip connections.
        For this parameter to have an impact, there has to be at least one skip connection in the model.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseFCN"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        rank,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        use_skip_connection=False,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)

        self.rank = rank
        self.data_format = data_format
        self.activation = activation
        self.use_bias = use_bias
        self.merge_layer = merge_layer
        
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint

        # set iterable class variables for the model
        self.set_vars(
            filters=filters, 
            kernel_size=kernel_size, 
            strides=strides,
            padding=padding,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=use_skip_connection
        )

        # check `padding` mode
        if (self.rank != 1) & any(item != "same" for item in self.padding):
            raise ValueError(f"For `rank={self.rank}`, `'same'` is the only valid padding mode, received `padding={self.padding}`.")
        
        if (self.rank == 1) & any(item not in ["same", "causal"] for item in self.padding):
            raise ValueError(f"Valid padding modes for `rank=1` are `'same'` and `'causal'`, received `padding={self.padding}`.")

        self.set_encoder_layers()
        self.bottleneck = bottleneck
        self.set_decoder_layers()

        # cache build shapes
        self.global_build_shapes_dict = None        

    def set_encoder_layers(self):
        self.encoder_layers = [
            getattr(import_module(name="...blocks.autoencoder", package=__package__), f"Encoder{self.rank}D")(
                filters=f,
                kernel_size=k,
                strides=s,
                padding=p,
                data_format=self.data_format,
                dilation_rate=d,
                groups=g,
                use_skip_connection=skip,
                activation=self.activation,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                activity_regularizer=self.activity_regularizer,
                name=f"Encoder_{i}"
            ) for i, (f, k, s, p, d, g, skip) in enumerate(
                zip(
                    self.filters,
                    self.kernel_size, 
                    self.strides,
                    self.padding, 
                    self.dilation_rate, 
                    self.groups, 
                    self.use_skip_connection
                )
            )
        ]
    
    def set_decoder_layers(self):
        self.decoder_layers = [
            getattr(import_module(name="...blocks.autoencoder", package=__package__), f"Decoder{self.rank}D")(
                filters=f,
                kernel_size=k,
                strides=s,
                padding=p,
                data_format=self.data_format,
                dilation_rate=d,
                groups=g,
                merge_layer=self.merge_layer, 
                activation=self.activation,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                activity_regularizer=self.activity_regularizer,
                name=f"Decoder_{len(self.filters) - i - 1}"
            ) for i, (f, k, s, p, d, g) in enumerate(
                zip(
                    list(reversed(self.filters)), 
                    list(reversed(self.kernel_size)),
                    list(reversed(self.strides)),
                    list(reversed(self.padding)),
                    list(reversed(self.dilation_rate)),
                    list(reversed(self.groups))
                )
            )
        ]

    def build(self, input_shape):
        """
        Build method of `BaseFNO`

        Builds all sub-modules and sets `self.built=True`.
        
        Parameters
        ----------
        input_shape : tuple
            Input shape

        """

        if self.built:
            return
        
        # cache build shapes
        self.global_build_shapes_dict = {}

        # build encoder layers
        output_shapes = []
        for layer, layer_uses_skip in zip(self.encoder_layers, self.use_skip_connection):
            # cache input shape
            self.global_build_shapes_dict.update({layer.name: {"input_shape": input_shape}})

            layer.build(input_shape=input_shape)

            # update input shape and append to output_shapes
            input_shape = layer.compute_output_shape(input_shape=input_shape)
            if layer_uses_skip:
                input_shape, input_shape_skip = input_shape
                output_shapes.append(input_shape_skip)
            else:
                output_shapes.append(None)

        # build bottleneck layers
        if self.bottleneck is not None:
            # cache input shape
            self.global_build_shapes_dict.update({self.bottleneck.name: {"input_shape": input_shape}})

            self.bottleneck.build(input_shape=input_shape)
            input_shape = self.bottleneck.compute_output_shape(input_shape=input_shape)

        # now build decoder layers
        output_shapes.reverse()

        for layer, input_shape_skip in zip(self.decoder_layers, output_shapes):
            # cache input shape
            self.global_build_shapes_dict.update({layer.name: {"input_shape": input_shape, "input_shape_skip": input_shape_skip}})

            layer.build(input_shape=input_shape, input_shape_skip=input_shape_skip)
            input_shape = layer.compute_output_shape(input_shape=input_shape, input_shape_skip=input_shape_skip)

        self.built = True

    def call(self, inputs):
        """
        
        Parameters
        ----------
        inputs : KerasTensor
            Input to BaseFCN

        Returns
        -------
        outputs : KerasTensor
            Output of BaseFCN

        """

        skip = []
        # forward path through encoder
        x = inputs
        for layer, layer_uses_skip in zip(self.encoder_layers, self.use_skip_connection):
            x = layer(x)
            if layer_uses_skip:
                x, x_skip = x
                skip.append(x_skip)
            else:
                skip.append(None)

        # apply bottlebeck layer
        if self.bottleneck is not None:
            x = self.bottleneck(x)

        # forward path through decoder
        for i, layer in enumerate(self.decoder_layers):
            x = layer(x, skip=skip[len(self.decoder_layers) - i - 1])

        return x
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseFCN`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """

        skip = []
        for layer, layer_uses_skip in zip(self.encoder_layers, self.use_skip_connection):
            if layer_uses_skip:
                
                input_shape, skip_shape = layer.compute_output_shape(input_shape=input_shape)
                skip.append(skip_shape)
            else:
                input_shape = layer.compute_output_shape(input_shape=input_shape)
                skip.append(None)

        # apply bottlebeck layer
        if self.bottleneck is not None:
            input_shape = self.bottleneck.compute_output_shape(input_shape=input_shape)

        # forward path through decoder
        for i, layer in enumerate(self.decoder_layers):
            input_shape = layer.compute_output_shape(input_shape=input_shape, input_shape_skip=skip[len(self.decoder_layers) - i - 1])

        return input_shape

    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.
            
        """

        config: dict = super().get_config()
        config.update({
            "filters": self.filters,
            "kernel_size": self.kernel_size,
            "padding": self.padding,
            "data_format": self.data_format,
            "dilation_rate": self.dilation_rate,
            "groups": self.groups,
            "use_skip_connection": self.use_skip_connection,
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "activation": saving.serialize_keras_object(self.activation),
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint)
        })
        
        if self.bottleneck is not None:
            config.update({"bottleneck": saving.serialize_keras_object(self.bottleneck)})
        return config
    
    @classmethod
    def from_config(cls, config):
        """
        Necessary for Keras deserialization

        Parameters
        ----------
        cls : BaseFCN
            The `BaseFCN` class.
        config : dict
            Dictionary with the layer configuration.

        Returns
        -------
        cls : BaseFCN
            Instance of `BaseFCN` from `config`.
            
        """

        activation_cfg = config.pop("activation")
        merge_layer_cfg = config.pop("merge_layer")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")
        kernel_constraints_cfg = config.pop("kernel_constraint")
        bias_constraints_cfg = config.pop("bias_constraint")
        bottleneck_cfg = config.pop("bottleneck", None)

        # now update with deserialized version
        config.update({
            "activation": saving.deserialize_keras_object(activation_cfg),
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraints_cfg),
            "bias_constraint": constraints.deserialize(bias_constraints_cfg)
        })
        if bottleneck_cfg:
            config.update({"bottleneck": saving.deserialize_keras_object(bottleneck_cfg)})

        return cls(**config)
    
    def get_build_config(self) -> dict:
        return self.global_build_shapes_dict
    
    def build_from_config(self, config):
        for layer in self.layers:
            try:
                layer.build(**config[layer.name])
            except ValueError:
                # layer is already build
                pass
            except KeyError:
                # layer has not input shape, e.g., activation layer like ReLU
                pass

        self.built = True


class BaseUnet(BaseFCN):
    """
    Base class of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    Convolutional autoencoder *with* skip-connections

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseUnet`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional
        Padding for all convolutional layers in the model.
        If `rank=1`, `padding` may be either `"same"` or `"causal"`, `rank>1` enforces `padding="same"`
        to maintain deterministic shapes throughout the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseUnet"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(self,
        rank,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1, 
        groups=1, 
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=True,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )

    def get_config(self) -> dict:
        config = super().get_config()
        config.pop("use_skip_connection", None)  # This should always be `True` for UNet

        return config


### MODELS THAT USE SMOOTH UPSAMPLING LAYER ###
class BaseSmoothFCN(BaseFCN):
    """
    Base class of FullyConvolutionalNetwork (FCN)

    Convolutional autoencoder *without* skip-connections

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseFCN`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional
        Padding for encoder blocks. 
        Decoder blocks will mirror the `padding` of the encoder blocks.
        If `rank=1`, `padding` may be either `"same"` or `"causal"`, `rank>1` enforces `padding="same"`
        to maintain deterministic shapes throughout the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    use_skip_connection : bool | list, optional
        Whether to use skip connections (basically transform FCN to a Unet).
        Can be defined per layer, i.e., 
        `use_skip_connection=[True, False, False]` results in a skip connection on the highest level only.
        Defaults to `False`.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the (optional) information from skip connections.
        For this parameter to have an impact, there has to be at least one skip connection in the model.
        Defaults to `"concatenate"`.
    interpolation : str, optional
        Interpolation to use in `SmoothUpSampling` layer.
        This option is only valid for `rank=2` and is ignored otherwise.
        Defaults to `"nearest"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseFCN"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """

    def __init__(
        self,
        rank,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1,
        groups=1,
        use_skip_connection=False,
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        interpolation="nearest",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        self.interpolation = interpolation
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=use_skip_connection,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            merge_layer=merge_layer,
            **kwargs
        )


    def set_decoder_layers(self):
        self.decoder_layers = [
            getattr(import_module(name="...blocks.autoencoder", package=__package__), f"SmoothDecoder{self.rank}D")(
                filters=f,
                kernel_size=k,
                strides=s,
                padding=p,
                data_format=self.data_format,
                dilation_rate=d,
                groups=g,
                merge_layer=self.merge_layer, 
                interpolation=self.interpolation,
                activation=self.activation,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                activity_regularizer=self.activity_regularizer,
                name=f"Decoder_{len(self.filters) - i - 1}"
            ) for i, (f, k, s, p, d, g) in enumerate(
                zip(
                    list(reversed(self.filters)), 
                    list(reversed(self.kernel_size)),
                    list(reversed(self.strides)),
                    list(reversed(self.padding)),
                    list(reversed(self.dilation_rate)),
                    list(reversed(self.groups))
                )
            )
        ]

    def get_config(self):
        """
        Necessary for Keras serialization

        Returns
        -------
        config : dict
            Dictionary with the layer configuration.
            
        """

        config = super().get_config()
        config.update({"interpolation": self.interpolation})

        return config
    

class BaseSmoothUnet(BaseSmoothFCN):
    """
    Base class of Unet, cf. [Ronneberger et al.](https://arxiv.org/abs/1505.04597)

    Convolutional autoencoder *with* skip-connections

    Upsampling is performed via a stack of image upsampling (nearest neighbor or equivalent, define as `interpolation`) and a convolution.

    Parameters
    ----------
    rank : int {1, 2, 3}
        Rank of `BaseUnet`. Must be within {1, 2, 3}.
    filters : int | list | tuple
        Number of filters. 
        The model will be build symmetrically, meaning that only the number of encoder filters is defined,
        and the decoder will mirror it.
    kernel_size : int | list | tuple, optional
        Kernel size for encoder blocks. 
        An `int` results in a global `kernel_size`, a `list` allows to define the `kernel_size` per layer.
        Decoder blocks will mirror the `kernel_size` of the encoder blocks.
        Defaults to 5.
    strides : int | list | tuple, optional
        Strides for encoder blocks. 
        An `int` results in a global value for `strides`, a `list` allows to define the `strides` per layer.
        Decoder blocks will mirror the `strides` of the encoder blocks.
        Defaults to 1.
    padding : str, optional
        Padding for all convolutional layers in the model.
        If `rank=1`, `padding` may be either `"same"` or `"causal"`, `rank>1` enforces `padding="same"`
        to maintain deterministic shapes throughout the model.
        Defaults to `"same"`.
    data_format : str, optional {`"channels_first"`, `"channels_last"`}
        Data format for the convolution operations.
        Defaults to `"channels_last"`.
    dilation_rate : int | list | tuple, optional
        Dilation rates for encoder blocks. 
        An `int` results in a global value for `dilation_rate`, a `list` allows to define the `dilation_rate` per layer.
        Decoder blocks will mirror the `dilation_rate` of the encoder blocks.
        Defaults to 1.
    groups : int, optional
        Number of convolutional groups for encoder blocks. 
        An `int` results in a global value for `groups`, a `list` allows to define the `groups` per layer.
        Decoder blocks will mirror the `groups` of the encoder blocks.
        Defaults to 1.
    activation : str | keras.activations.Activation | keras.layers.Layer, optional
        Global activation function.
        Can be either a `str`, a `keras.activations.Activation`, or a `keras.layers.Layer`.
        Defaults to `"relu"`.
    use_bias : bool, optional
        If `True`, all layers use a bias.
        Defaults to `True`.
    bottleneck : keras.layers.Layer | keras.models.Model, optional
        An optional `keras.layer.Layer` or `keras.models.Model` that is placed in the bottleneck of the model.
        Defaults to `None`.
    merge_layer : str | keras.layers.Layer, optional {`"concatenate"`, `"average"`, `"maximum"`, `"minimum"`, `"add"`, `"subtract"`, `"multiply"`, `"dot"`}
        Layer to merge the forward information with the information from skip connections.
        Defaults to `"concatenate"`.
    interpolation : str, optional
        Interpolation to use in `SmoothUpSampling` layer.
        This option is only valid for `rank=2` and is ignored otherwise.
        Defaults to `"nearest"`.
    kernel_initializer : str | keras.initializers.Initializer, optional
        Kernel initializer.
        Defaults to `"he_normal"`.
    bias_initializer : str | keras.initializers.Initializer, optional
        Bias initializer.
        Defaults to `"zeros"`.
    kernel_regularizer : str | keras.regularizers.Regularizer, optional
        Kernel regularizer.
        Defaults to `None`.
    bias_regularizer : str | keras.regularizers.Regularizer, optional
        Bias regularizer.
        Defaults to `None`.
    kernel_constraint : str | keras.constraints.Constraint, optional
        Kernel constraint.
        Defaults to `None`.
    bias_constraint : str | keras.constraints.Constraint, optional
        Bias constraint.
        Defaults to `None`.
    name : str, optional
        Name of the model.
        If `None`, `name` is automatically inherited from the class name `"BaseUnet"`.
        Defaults to `None`.

    Notes
    -----
    The `filters` argument steers the depth of the model and the layout of the encoder- and decoder blocks on each level:
    A list containing tuples results in multiple convolutions on the respective level, e.g.,
    `filters=[(8, 8), 16, 32]` will return a model with a depth of 3, 
    where the first encoder (and last decoder) block has two consecutive convolutions with `filters=8`,
    and the remaining two levels have only a single convolution with `filters=16` and `filters=32`, respectively.

    Downsampling is realized using a strided convolution as proposed by [Springenberger et al.](https://arxiv.org/abs/1412.6806)

    """
    def __init__(self,
        rank,
        filters=[8, 16, 32],
        kernel_size=5,
        strides=1,
        padding="same",
        data_format="channels_last",
        dilation_rate=1, 
        groups=1, 
        activation="relu",
        use_bias=True,
        bottleneck=None,
        merge_layer="concatenate",
        interpolation="nearest",
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=rank,
            filters=filters,
            kernel_size=kernel_size,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilation_rate=dilation_rate,
            groups=groups,
            use_skip_connection=True,
            activation=activation,
            use_bias=use_bias,
            bottleneck=bottleneck,
            merge_layer=merge_layer,
            interpolation=interpolation,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            name=name,
            **kwargs
        )


from keras import layers
from keras import activations
from keras import initializers
from keras import regularizers
from keras import constraints
from keras import saving
from ...layers.wrapper import Residual
from importlib import import_module


class BaseFNO(layers.Layer):
    def __init__(
        self,
        rank,
        filters,
        modes,
        activation="gelu",
        merge_layer="add",
        data_format="channels_last",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(name=name, **kwargs)
        self.rank = rank
        self.filters = filters
        self.modes = modes
        self.merge_layer = merge_layer
        self.data_format = data_format
        self.use_bias = use_bias
        self.kernel_initializer = kernel_initializer
        self.bias_initializer = bias_initializer
        self.kernel_constraint = kernel_constraint
        self.bias_constraint = bias_constraint
        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.forward = Residual(
            layer=getattr(import_module(name=".spectral_conv", package=__package__), f"SpectralConv{self.rank}D")(
                filters=self.filters,
                modes=self.modes,
                data_format=self.data_format,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                name="spectral_conv",
                **kwargs
            ),
            merge_layer=self.merge_layer,
            residual_layer=getattr(import_module(name="...layers.conv", package=__package__), f"Conv{self.rank}D")(
                filters=self.filters,
                kernel_size=1,
                data_format=self.data_format,
                use_bias=self.use_bias,
                kernel_initializer=self.kernel_initializer,
                bias_initializer=self.bias_initializer,
                kernel_constraint=self.kernel_constraint,
                bias_constraint=self.bias_constraint,
                kernel_regularizer=self.kernel_regularizer,
                bias_regularizer=self.bias_regularizer,
                name="bypass_conv"
            )
        )
        
        self.activation = activations.get(activation)

    def build(self, input_shape):
        if self.built:
            return
        
        self.forward.build(input_shape=input_shape)

        # inherit input spec from `spectral_conv` layer
        self.input_spec = self.forward.layer.input_spec
        
        self.built = True

    def call(self, inputs):
        x = self.forward(inputs)
        x = self.activation(x)

        return x
    
    def compute_output_shape(self, input_shape):
        """
        Compute output shape of `BaseNeuralOperator`

        Parameters
        ----------
        input_shape : tuple
            Input shape.

        Returns
        -------
        output_shape : tuple
            Output shape.

        """
        
        return self.forward.compute_output_shape(input_shape=input_shape)        
    
    def get_config(self):
        config = super().get_config()
        config.update({
            "filters": self.filters,
            "modes": self.modes,
            "activation": saving.serialize_keras_object(self.activation),
            "merge_layer": saving.serialize_keras_object(self.merge_layer),
            "data_format": self.data_format,
            "use_bias": self.use_bias,
            "kernel_initializer": initializers.serialize(self.kernel_initializer),
            "bias_initializer": initializers.serialize(self.bias_initializer),
            "kernel_constraint": constraints.serialize(self.kernel_constraint),
            "bias_constraint": constraints.serialize(self.bias_constraint),
            "kernel_regularizer": regularizers.serialize(self.kernel_regularizer),
            "bias_regularizer": regularizers.serialize(self.bias_regularizer)
        })

        return config
    
    @classmethod
    def from_config(cls, config):
        activation_cfg = config.pop("activation")
        merge_layer_cfg = config.pop("merge_layer")
        kernel_initializer_cfg = config.pop("kernel_initializer")
        bias_initializer_cfg = config.pop("bias_initializer")
        kernel_constraint_cfg = config.pop("kernel_constraint")
        bias_constraint_cfg = config.pop("bias_constraint")
        kernel_regularizer_cfg = config.pop("kernel_regularizer")
        bias_regularizer_cfg = config.pop("bias_regularizer")

        config.update({
            "activation": saving.deserialize_keras_object(activation_cfg),
            "merge_layer": saving.deserialize_keras_object(merge_layer_cfg),
            "kernel_initializer": initializers.deserialize(kernel_initializer_cfg),
            "bias_initializer": initializers.deserialize(bias_initializer_cfg),
            "kernel_constraint": constraints.deserialize(kernel_constraint_cfg),
            "bias_constraint": constraints.deserialize(bias_constraint_cfg),
            "kernel_regularizer": regularizers.deserialize(kernel_regularizer_cfg),
            "bias_regularizer": regularizers.deserialize(bias_regularizer_cfg)
        })

        return cls(**config)
    


from keras import saving
from keras import layers
from keras.src import ops


@saving.register_keras_serializable(package="Kerex.Blocks.Attention", name="MultiHeadAttention")
class MultiHeadAttention(layers.MultiHeadAttention):
    """
    
    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    key_dim : int
        Size of each attention head for query and key.
    value_dim : int, optional
        Size of each attention head for value.
        If this is `None`, it is set to `key_dim`.
        Defaults to `None`.
    dropout : float, optional
        Dropout probability. Defaults to 0.
    use_bias : bool, optional
        If this is set, a bias is ued. Defaults to `True`.
    output_shape : tuple, optional
        The expected shape of an output tensor, besides the batch and sequence dims.
        If not specified, projects back to the query feature dim (the query input's last dimension).
        Defaults to `None`.
    attention_axes : int, optional
        Axes over which the attention is applied.
        `None` means attention over all axes, but batch, heads, and features.
        Defaults to `None`.
    kernel_initializer : str, optional
        Initializer for dense layer kernels. Defaults to `"glorot_uniform"`.
    bias_initializer : str, optional
        Initializer for dense layer biases. Defaults to `"zeros"`.
    kernel_regularizer : str, optional
        Regularizer for dense layer kernels. Defaults to `None`.
    bias_regularizer : str, optional
        Regularizer for dense layer biases. Defaults to `None`.
    activity_regularizer : str, optional
        Regularizer for dense layer activity. Defaults to `None`.
    kernel_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    bias_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    seed : int, optional
        Optional integer to seed the dropout layer. Defaults to `None`

    """
    
    def _compute_attention_mask(
        self,
        query,
        value,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        use_causal_mask=False,
        prefix: int = 0,
        prediction_step: int = 0
    ):
        """
        Computes the attention mask, using the Keras masks of the inputs.

        * The `query`'s mask is reshaped from [B, T] to [B, T, 1].
        * The `value`'s mask is reshaped from [B, S] to [B, 1, S].
        * The `key`'s mask is reshaped from [B, S] to [B, 1, S]. The `key`'s
          mask is ignored if `key` is `None` or if `key is value`.
        * If `use_causal_mask=True`, then the causal mask is computed. Its shape
          is [1, T, S].

        All defined masks are merged using a logical AND operation (`&`).

        In general, if the `query` and `value` are masked, then there is no need
        to define the `attention_mask`.

        Parameters
        ----------
        query : KerasTensor
            Projected query tensor of shape `(B, T, N, key_dim)`.
        key : KerasTensor
            Projected key tensor of shape `(B, T, N, key_dim)`.
        value : KerasTensor
            Projected value tensor of shape `(B, T, N, value_dim)`.
        attention_mask : KerasTensor
            A boolean mask of shape `(B, T, S)`, that prevents attention to certain positions.
        use_causal_mask : bool
            A boolean to indicate whether to apply a causal mask to prevent tokens from attending to future tokens (e.g., used in a decoder Transformer).

        Returns
        -------
        attention_mask : KerasTensor
            A boolean mask of shape `(B, T, S)`, that prevents attention to certain positions, 
            based on the Keras masks of the `query`, `key`, `value`, and `attention_mask` tensors, 
            and the causal mask if `use_causal_mask=True`.

        """

        auto_mask = None
        if query_mask is not None:
            query_mask = ops.cast(query_mask, "bool")  # defensive casting
            # B = batch size, T = max query length
            auto_mask = ops.expand_dims(query_mask, -1)  # shape is [B, T, 1]
        if value_mask is not None:
            value_mask = ops.cast(value_mask, "bool")  # defensive casting
            # B = batch size, S == max value length
            mask = ops.expand_dims(value_mask, -2)  # shape is [B, 1, S]
            auto_mask = mask if auto_mask is None else auto_mask & mask
        if key_mask is not None:
            key_mask = ops.cast(key_mask, "bool")  # defensive casting
            # B == batch size, S == max key length == max value length
            mask = ops.expand_dims(key_mask, -2)  # shape is [B, 1, S]
            auto_mask = mask if auto_mask is None else auto_mask & mask
        if use_causal_mask:
            # the shape of the causal mask is [1, T, S]
            mask = self._compute_causal_mask(query, value, prefix, prediction_step)
            auto_mask = mask if auto_mask is None else auto_mask & mask
        if auto_mask is not None:
            # merge attention_mask & automatic mask, to shape [B, T, S]
            attention_mask = (
                auto_mask
                if attention_mask is None
                else ops.cast(attention_mask, bool) & auto_mask
            )
        return attention_mask

    def _compute_causal_mask(self, query, value=None, prefix: int = 0, prediction_step: int = 0):
        """
        Computes a causal mask (e.g., for masked self-attention layers).

        For example, if query and value both contain sequences of length 4,
        this function returns a boolean tensor equal to:

        ```
        [[[True,  False, False, False],
          [True,  True,  False, False],
          [True,  True,  True,  False],
          [True,  True,  True,  True]]]
        ```

        Parameters
        ----------
        query : KerasTensor
            A query tensor of shape `(B, T, ...)`.
        value : KerasTensor
            A value tensor of shape `(B, S, ...)` (optional, defaults to query).

        Returns
        -------
        mask : KerasTensor
            A boolean tensor of shape `(1, T, S)` containing a lower triangular matrix of shape `(T, S)`.

        """

        q_seq_length = prediction_step or ops.shape(query)[1]
        v_seq_length = ops.shape(query)[1] if value is None else ops.shape(value)[1]
        ones_mask = ops.ones((1, q_seq_length, v_seq_length), dtype="int32")
        row_index = ops.cumsum(ones_mask, axis=-2)
        col_index = ops.cumsum(ones_mask, axis=-1)

        prefix = ops.full((1, q_seq_length, v_seq_length), prefix, dtype="int32")
        offset_mask = ops.greater_equal(prefix, col_index)
        offset_mask = ops.multiply(prefix, ops.cast(offset_mask, dtype="int32"))
        row_index = ops.add(row_index, offset_mask)

        mask = ops.greater_equal(row_index, col_index)
        return ops.pad(mask, pad_width=((0, 0), (0, v_seq_length - q_seq_length), (0, 0)))
    
    def call(
        self,
        query,
        value,
        key=None,
        query_mask=None,
        value_mask=None,
        key_mask=None,
        attention_mask=None,
        return_attention_scores=False,
        return_attention_mask=False, 
        training=None,
        use_causal_mask=False,
        prefix=0,
        prediction_step=0
    ):
        if key is None:
            key = value

        attention_mask = self._compute_attention_mask(
            query,
            value,
            query_mask=query_mask,
            value_mask=value_mask,
            key_mask=key_mask,
            attention_mask=attention_mask,
            use_causal_mask=use_causal_mask,
            prefix=prefix,
            prediction_step=prediction_step
        )

        #   N = `num_attention_heads`
        #   H = `size_per_head`
        # `query` = [B, T, N ,H]
        query = self._query_dense(query)

        # `key` = [B, S, N, H]
        key = self._key_dense(key)

        # `value` = [B, S, N, H]
        value = self._value_dense(value)

        attention_output, attention_scores = self._compute_attention(
            query, key, value, attention_mask, training
        )
        attention_output = self._output_dense(attention_output)

        if return_attention_scores:
            return attention_output, attention_scores
        if return_attention_mask:
            return attention_output, attention_mask
        return attention_output


@saving.register_keras_serializable(package="Kerex.Blocks.Attention", name="BaseAttention")
class BaseAttention(layers.Layer):
    """
    Base multihead-attention layer as described in the paper "Attention is all you Need" [Vaswani et al., 2017](https://arxiv.org/abs/1706.03762).
    
    This BaseAttention class implements the Attention block, consisting in
    - Multihead-Attention layer, and
    - LayerNorm,
    and serves as a BaseLayer to inherit from.

    Parameters
    ----------
    num_heads : int
        Number of attention heads.
    key_dim : int
        Size of each attention head for query and key.
    value_dim : int, optional
        Size of each attention head for value.
        If this is `None`, it is set to `key_dim`.
        Defaults to `None`.
    dropout : float, optional
        Dropout probability. Defaults to 0.
    use_bias : bool, optional
        If this is set, a bias is ued. Defaults to `True`.
    output_shape : tuple, optional
        The expected shape of an output tensor, besides the batch and sequence dims.
        If not specified, projects back to the query feature dim (the query input's last dimension).
        Defaults to `None`.
    attention_axes : int, optional
        Axes over which the attention is applied.
        `None` means attention over all axes, but batch, heads, and features.
        Defaults to `None`.
    kernel_initializer : str, optional
        Initializer for dense layer kernels. Defaults to `"glorot_uniform"`.
    bias_initializer : str, optional
        Initializer for dense layer biases. Defaults to `"zeros"`.
    kernel_regularizer : str, optional
        Regularizer for dense layer kernels. Defaults to `None`.
    bias_regularizer : str, optional
        Regularizer for dense layer biases. Defaults to `None`.
    activity_regularizer : str, optional
        Regularizer for dense layer activity. Defaults to `None`.
    kernel_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    bias_constraint : str, optional
        Constraint for dense layer kernels. Defaults to `None`.
    seed : int, optional
        Optional integer to seed the dropout layer. Defaults to `None`

    Notes
    -----
    With flash attention (Keras>3.6.0), the dropout probability has to be 0.

    """

    def __init__(
            self, 
            num_heads, 
            key_dim, 
            value_dim=None, 
            dropout=0, 
            use_bias=True,  
            output_shape=None, 
            attention_axes=None, 
            kernel_initializer="glorot_uniform", 
            bias_initializer="zeros", 
            kernel_regularizer=None, 
            bias_regularizer=None, 
            activity_regularizer=None, 
            kernel_constraint=None, 
            bias_constraint=None, 
            seed=None,
            **kwargs
        ):
        super().__init__()

        self.mha = MultiHeadAttention(num_heads, key_dim, value_dim, dropout, use_bias, output_shape, attention_axes, kernel_initializer, bias_initializer, kernel_regularizer, bias_regularizer, activity_regularizer, kernel_constraint, bias_constraint, seed, **kwargs)
        self.layernorm = layers.LayerNormalization()
        self.add = layers.Add()

    def build(self, query_shape, value_shape, key_shape=None):
        """
        Build method of BaseAttention

        Parameters
        ----------
        query_shape : tuple
            Input shape of query tensor.
        value_shape : tuple
            Input shape of value tensor.
        key_shape : tuple, optional
            Input shape of key tensor. Defaults to `None`.

        """

        if self.built:
            return
        
        self.mha.build(query_shape=query_shape, value_shape=value_shape, key_shape=key_shape)
        self.layernorm.build(input_shape=query_shape)
        self.add.build(input_shape=[query_shape, query_shape])

        self.built = True

    def compute_output_shape(self, input_shape: tuple) -> tuple:
        """
        Compute output shape of layer

        Parameters
        ----------
        input_shape : tuple | list
            Shape of the input

        Returns
        -------
        output_shape : tuple
            Shape of the output (same as input!)

        """

        return tuple(input_shape)
    
    def call(self, *args):
        """
        Placeholder that ensures that the BaseAttention is not actually used.

        """

        raise NotImplementedError("Don't even think about calling this base layer!")



from keras import saving
from .base_spectral_conv import BaseSpectralConv


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv1D")
class SpectralConv1D(BaseSpectralConv):
    def __init__(
        self,
        filters,
        modes,
        data_format="channels_last",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=1,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv2D")
class SpectralConv2D(BaseSpectralConv):
    def __init__(
        self,
        filters,
        modes,
        data_format="channels_last",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=2,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )


@saving.register_keras_serializable(package="Kerex.Layers.FNO", name="SpectralConv3D")
class SpectralConv3D(BaseSpectralConv):
    def __init__(
        self,
        filters,
        modes,
        data_format="channels_last",
        use_bias=True,
        kernel_initializer="he_normal",
        bias_initializer="zeros",
        kernel_constraint=None,
        bias_constraint=None,
        kernel_regularizer=None,
        bias_regularizer=None,
        name=None,
        **kwargs
    ):
        super().__init__(
            rank=3,
            filters=filters,
            modes=modes,
            data_format=data_format,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            name=name,
            **kwargs
        )
